/*  fimview v0.1 Copyright (C) 2003, 2004 tomasliq <tom@eggdrop.ch>
 *  
 *  main.c -- main file
 *
 *  This file is part of fimview.
 *  For more information visit: http://www.eggdrop.ch/projects/fimview/
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


/*
 * includes
 */

/* ANSI C */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>

/* X11 */
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/shape.h>
#include <X11/keysym.h>

/* imlib */
#include <Imlib2.h>

/* for stat() */
#include <sys/types.h>
#include <sys/stat.h>

/* for realpath() */
#include <sys/param.h>

/* for opendir() etc */
#include <dirent.h>


/*
 * macros
 */

/* name */
#define FIM_CLASS "FimView"
#define FIM_NAME "fimview"

/* number mouse buttons */
#define FIM_BUTTONS 5

/* this is dependend how many RAM you have default is 20 MB, increase it if 
 * you often load files that are bigger (uncompressed!) */
#define FIM_CACHE 20 * 1024 * 1024 

/* default window width and height in pixels */
#define FIM_WINWIDTH 640
#define FIM_WINHEIGHT 480


/*
 * structures
 */

struct fim_files
{
	char *name; /* full path name */
	struct fim_files *next; /* next file or NULL */
	struct fim_files *prev; /* previous file or NULL */
};

struct fim_key
{
	char antialiasing;
};

struct fim_x
{
	Display *dpy; /* the display */
	Window win; /* the main window */
	Atom atom_delete; /* this atom enables closing with a windowmanager */
	int screen_x, screen_y; /* screen width and height */
	int old_x, old_y; /* old mouse coordinates */
	int mouse_x, mouse_y; /* current mouse coordinates */
	char button_pressed[FIM_BUTTONS]; /* which buttons are presed? */
	char button_released[FIM_BUTTONS]; /* which buttons were released? */
	struct fim_files *files; /* list with the file names */
	struct fim_files *cur; /* the current file */
	Imlib_Updates update; /* imlib update */
	int win_w, win_h; /* window width, height */
	struct fim_key key; /* stores pressed keys and options */
	int done; /* the program exists, if this is set to 1 */
};


/*
 * prototypes
 */

int main(int argc, char *argv[]);
void draw(struct fim_x *x);
void keyhandler(struct fim_x *x, char mode, unsigned int keycode);
void mainloop(struct fim_x *x);
void init_struct(struct fim_x *x);
void setup_x(struct fim_x *x);
void loadfile(struct fim_x *x, char *file);
void loaddirectory(struct fim_x *x, char *dir);
void loadfilelist(struct fim_x *x, int argc, char *argv[]);


/*
 * functions
 */

void draw(struct fim_x *x)
{
	int i;
	Imlib_Updates cup;
	Imlib_Image buf, img;

	x->update = imlib_updates_merge_for_rendering(x->update, \
			x->win_w, x->win_h);

	for (cup=x->update ; cup ; cup=imlib_updates_get_next(cup))
	{
		int u_x, u_y, u_w, u_h, img_w, img_h;

		imlib_updates_get_coordinates(cup, &u_x, &u_y, &u_w, &u_h);

		buf = imlib_create_image(u_w, u_h);

		imlib_context_set_blend(1);
		
		img = imlib_load_image(x->cur->name);
		imlib_context_set_image(img);
		img_w = imlib_image_get_width();
		img_h = imlib_image_get_height();

		imlib_context_set_image(buf);
		if (!img)
		{
			fprintf(stderr, "cannot load image %s\n", \
					x->cur->name);
		}
		else
		{
			imlib_blend_image_onto_image(img, 0, 0, 0, \
				img_w, img_h, -u_x, -u_y, x->win_w, x->win_h);
			imlib_context_set_image(img);
			imlib_free_image();
		}

		imlib_context_set_blend(0);
		imlib_context_set_image(buf);
		imlib_render_image_on_drawable(u_x, u_y);
		imlib_free_image();
	}

	if (x->update)
	{
		imlib_updates_free(x->update);
		x->update = imlib_updates_init();
	}

	/* some mouse settings */

	for (i=0;i<FIM_BUTTONS;i++)
		x->button_released[i] = 0;

	x->old_x = x->mouse_x;
	x->old_y = x->mouse_y;

	/* correct mouse behaviour (currently not used) */

	/* vertical */
	if (x->mouse_x == x->screen_x-1)
		XWarpPointer(x->dpy, None, DefaultRootWindow(x->dpy), 0, 0, \
				0, 0, 1, x->mouse_y);
	if (x->mouse_x == 0)
		XWarpPointer(x->dpy, None, DefaultRootWindow(x->dpy), 0, 0, \
				0, 0, x->screen_x-2, x->mouse_y);

	/* horizontal */
	if (x->mouse_y == x->screen_y-1)
		XWarpPointer(x->dpy, None, DefaultRootWindow(x->dpy), 0, 0, \
				0, 0, x->mouse_x, 1);
	if (x->mouse_y == 0)
		XWarpPointer(x->dpy, None, DefaultRootWindow(x->dpy), 0, 0, \
				0, 0, x->mouse_x, x->screen_y-2);
}

void keyhandler(struct fim_x *x, char mode, unsigned int keycode)
{
	if (mode == 1 && keycode == XKeysymToKeycode(x->dpy, XK_A))
	{
		/* toggle "antialiasing" */
		if (x->key.antialiasing)
		{
			x->key.antialiasing = 0;
			imlib_context_set_anti_alias(0);
		}
		else
		{
			x->key.antialiasing = 1;
			imlib_context_set_anti_alias(1);
		}

		/* redraw */
		x->update = imlib_update_append_rect(x->update, 0, 0, \
				x->win_w, x->win_h);
	}

	if (mode == 1 && keycode == XKeysymToKeycode(x->dpy, XK_N))
	{
		/* next picture */
		if (x->cur->next != NULL)
			x->cur = x->cur->next;

		/* redraw */
		x->update = imlib_update_append_rect(x->update, 0, 0, \
				x->win_w, x->win_h);
	}
	
	if (mode == 1 && keycode == XKeysymToKeycode(x->dpy, XK_P))
	{
		/* previous picture */
		if (x->cur->prev->prev != NULL)
			x->cur = x->cur->prev;

		/* redraw */
		x->update = imlib_update_append_rect(x->update, 0, 0, \
				x->win_w, x->win_h);
	}

	if (mode == 1 && keycode == XKeysymToKeycode(x->dpy, XK_Q))
	{
		/* exit */
		x->done = 1;
	}
}


void mainloop(struct fim_x *x)
{
	XEvent event;
	
	while (!x->done)
	{
		while (XPending(x->dpy) > 0)
		{
			XNextEvent(x->dpy, &event);
			switch (event.type)
			{
				case Expose:
x->update = imlib_update_append_rect(x->update, event.xexpose.x, \
		event.xexpose.y, event.xexpose.width, event.xexpose.height);
					break;
				case ButtonPress:
x->mouse_x = event.xbutton.x_root;
x->mouse_y = event.xbutton.y_root;
if (event.xbutton.button < 6 && event.xbutton.button > 0)
{
	x->button_pressed[event.xbutton.button-1]=1;
}
break;
				case ButtonRelease:
x->mouse_x = event.xbutton.x_root;
x->mouse_y = event.xbutton.y_root;
if (event.xbutton.button < 6 && event.xbutton.button > 0)
{
	x->button_pressed[event.xbutton.button-1]=0;
	x->button_released[event.xbutton.button-1]=1;
}
break;
				case MotionNotify:
					x->mouse_x = event.xmotion.x_root;
					x->mouse_y = event.xmotion.y_root;
					break;
				case KeyPress:
					keyhandler(x, 1, event.xkey.keycode);
					break;
				case KeyRelease:
					keyhandler(x, 0, event.xkey.keycode);
					break;
				case ClientMessage:
if (event.xclient.data.l[0]==x->atom_delete)
	x->done=1; /* the user closed the window */
					break;
				case ConfigureNotify:
					x->win_w = event.xconfigure.width;
					x->win_h = event.xconfigure.height;
					break;
				default:
					break;
			}
		}
		draw(x);
		usleep(10000); /* don't overload the CPU ;-) */
	}
}
void init_struct(struct fim_x *x)
{
	/* initialize structure */
	x->mouse_x = x->mouse_y = 1;
	x->old_x = x->old_y = 1;
	x->files = (struct fim_files *)malloc(sizeof(struct fim_files));
	x->files->next = NULL;
	x->files->prev = NULL;
	x->win_w = FIM_WINWIDTH;
	x->win_h = FIM_WINHEIGHT;
	x->update = imlib_updates_init();
	x->key.antialiasing = 1;
	x->done = 0;
}

void setup_x(struct fim_x *x)
{
	XClassHint class_hint;
	int i;

	for (i=0;i<FIM_BUTTONS;i++)
		x->button_pressed[i] = x->button_released[i] = 0;

	/* connect to X */
	if ((x->dpy = XOpenDisplay(NULL)) == NULL)
	{
		fprintf(stderr, "Unable to open display\n");
		exit(2);
	}

	x->screen_x = DisplayWidth(x->dpy, DefaultScreen(x->dpy));
	x->screen_y = DisplayHeight(x->dpy, DefaultScreen(x->dpy));

	/* create the window */
	x->win = XCreateSimpleWindow(x->dpy, DefaultRootWindow(x->dpy), 0, 0, \
		FIM_WINWIDTH, FIM_WINHEIGHT, 0, 0, \
		WhitePixel(x->dpy, DefaultScreen(x->dpy)));

	/* set the window name */
	class_hint.res_name = FIM_NAME;
	class_hint.res_class = FIM_CLASS;
	XSetClassHint(x->dpy, x->win, &class_hint);
	XStoreName(x->dpy, x->win, FIM_NAME);

	/* make it possible to close the window with a window manager */
	x->atom_delete = XInternAtom(x->dpy, "WM_DELETE_WINDOW", True);
	if (x->atom_delete)
		XSetWMProtocols(x->dpy, x->win, &(x->atom_delete), 1);

	/* input methods */
	XSelectInput(x->dpy, x->win, ExposureMask | KeyPressMask | \
		KeyReleaseMask | ButtonPressMask | ButtonReleaseMask | \
		PointerMotionMask | StructureNotifyMask);
	
	/* show the window */
	XMapWindow(x->dpy, x->win);

	/* Imlib */
	imlib_set_cache_size(FIM_CACHE);
	imlib_context_set_display(x->dpy);
	imlib_context_set_visual(DefaultVisual(x->dpy, DefaultScreen(x->dpy)));
	imlib_context_set_colormap(DefaultColormap(x->dpy, \
				DefaultScreen(x->dpy)));
	imlib_context_set_drawable(x->win);
	imlib_context_set_anti_alias(1);
}

void loadfile(struct fim_x *x, char *file)
{
	Imlib_Image img;
	struct fim_files *p;
	
	img = imlib_load_image(file);
	
	if (!img)
	{
		fprintf(stderr, "cannot load image %s\n", file);
		return;
	}

	p = x->files;
	while (p->next != NULL)
		p = p->next;
	
	p->next = (struct fim_files *)malloc(sizeof(struct fim_files));

	p->next->name = (char *)malloc(strlen(file)+1);
	strcpy(p->next->name, file);
	p->next->next = NULL;
	p->next->prev = p;

	imlib_context_set_image(img);
	imlib_free_image();
}

void loaddirectory(struct fim_x *x, char *dir)
{
	DIR *d;
	struct dirent *dent;
	char *path;
	struct stat attribute;

	if ((d = opendir(dir)) == NULL)
	{
		fprintf(stderr, "cannot open directory %s\n", dir);
		return;
	}

	fprintf(stderr, "scaning %s\n", dir);

	while ((dent = readdir(d)) != NULL)
	{
		int len = strlen(dir)+strlen(dent->d_name)+1;
		unsigned long mode;

		if (!strcmp(".", dent->d_name))
			continue;
		if (!strcmp("..", dent->d_name))
			continue;

		path = (char *)malloc(len+1);
		snprintf(path, len+1, "%s/%s", dir, dent->d_name);

		if (lstat(path, &attribute) != 0)
		{
			fprintf(stderr, "cannot stat %s (errno = %d)\n", \
					path, errno);
			continue;
		}

		mode = attribute.st_mode & S_IFMT;
		
		if (mode == S_IFDIR || mode == S_IFLNK)
		{
			/* recursive! */
			loaddirectory(x, path);
		}
		else if (mode == S_IFREG)
		{
			loadfile(x, path);
		}
	}

	closedir(d);
}

void loadfilelist(struct fim_x *x, int argc, char *argv[])
{
	int i;
	struct stat attribute;
	char path[MAXPATHLEN];

	for(i=1;i<argc;i++)
	{
		/* resolve symlinks */
		realpath(argv[i], path);

		if (stat(path, &attribute) != 0)
		{
			fprintf(stderr, "cannot stat %s (errno = %d)\n", \
					path, errno);
			continue;
		}
		
		if (attribute.st_mode & S_IFDIR)
		{
			/* it's a direcory */
			loaddirectory(x, path);
		}
		else if (attribute.st_mode & S_IFREG)
		{
			loadfile(x, path);
		}
	}
}

int main(int argc, char *argv[])
{
	struct fim_x x;

	/* check usage */
	if (argc < 2)
	{
		fprintf(stderr, "Usage: %s <file/directory> [...]\n", argv[0]);
		return 1;
	}

	init_struct(&x);
	loadfilelist(&x, argc, argv);
	if (x.files->next == NULL)
	{
		fprintf(stderr, "no images found\n");
		return 2;
	}
	x.cur = x.files->next;
	setup_x(&x);
	mainloop(&x);

	return 0;
}
